@@ -15,30 +15,28 @@ const bench = common.createBenchmark(main, {
 });
 
 function main({ n, len, type }) {
-  var arr = Array(len);
-  var i, opts;
+  let arr = Array(len);
+  let opts;
 
   switch (type) {
     case 'denseArray_showHidden':
       opts = { showHidden: true };
       arr = arr.fill('denseArray');
       break;
-      // For testing, if supplied with an empty type, default to denseArray.
-    case '':
     case 'denseArray':
       arr = arr.fill('denseArray');
       break;
     case 'sparseArray':
       break;
     case 'mixedArray':
-      for (i = 0; i < n; i += 2)
+      for (let i = 0; i < n; i += 2)
         arr[i] = i;
       break;
     default:
       throw new Error(`Unsupported type ${type}`);
   }
   bench.start();
-  for (i = 0; i < n; i++) {
+  for (let i = 0; i < n; i++) {
     util.inspect(arr, opts);
   }
   bench.end(n);
