@@ -24,8 +24,12 @@
 const {
   Array,
   ArrayIsArray,
+  ArrayPrototypePush,
+  FunctionPrototypeBind,
+  FunctionPrototypeCall,
   ObjectDefineProperty,
   ObjectSetPrototypeOf,
+  ReflectApply,
 } = primordials;
 
 const errors = require('internal/errors');
@@ -36,12 +40,12 @@ const {
 } = require('internal/dgram');
 const { guessHandleType } = internalBinding('util');
 const {
+  ERR_BUFFER_OUT_OF_BOUNDS,
   ERR_INVALID_ARG_TYPE,
   ERR_MISSING_ARGS,
   ERR_SOCKET_ALREADY_BOUND,
   ERR_SOCKET_BAD_BUFFER_SIZE,
   ERR_SOCKET_BUFFER_SIZE,
-  ERR_SOCKET_CANNOT_SEND,
   ERR_SOCKET_DGRAM_IS_CONNECTED,
   ERR_SOCKET_DGRAM_NOT_CONNECTED,
   ERR_SOCKET_DGRAM_NOT_RUNNING,
@@ -49,6 +53,7 @@ const {
 } = errors.codes;
 const {
   isInt32,
+  validateAbortSignal,
   validateString,
   validateNumber,
   validatePort,
@@ -69,6 +74,9 @@ const {
   SendWrap
 } = internalBinding('udp_wrap');
 
+const dc = require('diagnostics_channel');
+const udpSocketChannel = dc.channel('udp.socket');
+
 const BIND_STATE_UNBOUND = 0;
 const BIND_STATE_BINDING = 1;
 const BIND_STATE_BOUND = 2;
@@ -81,14 +89,18 @@ const RECV_BUFFER = true;
 const SEND_BUFFER = false;
 
 // Lazily loaded
-let cluster = null;
+let _cluster = null;
+function lazyLoadCluster() {
+  if (!_cluster) _cluster = require('cluster');
+  return _cluster;
+}
 
 const errnoException = errors.errnoException;
 const exceptionWithHostPort = errors.exceptionWithHostPort;
 
 
 function Socket(type, listener) {
-  EventEmitter.call(this);
+  FunctionPrototypeCall(EventEmitter, this);
   let lookup;
   let recvBufferSize;
   let sendBufferSize;
@@ -122,6 +134,25 @@ function Socket(type, listener) {
     recvBufferSize,
     sendBufferSize
   };
+
+  if (options?.signal !== undefined) {
+    const { signal } = options;
+    validateAbortSignal(signal, 'options.signal');
+    const onAborted = () => {
+      this.close();
+    };
+    if (signal.aborted) {
+      onAborted();
+    } else {
+      signal.addEventListener('abort', onAborted);
+      this.once('close', () => signal.removeEventListener('abort', onAborted));
+    }
+  }
+  if (udpSocketChannel.hasSubscribers) {
+    udpSocketChannel.publish({
+      socket: this,
+    });
+  }
 }
 ObjectSetPrototypeOf(Socket.prototype, EventEmitter.prototype);
 ObjectSetPrototypeOf(Socket, EventEmitter);
@@ -136,7 +167,7 @@ function startListening(socket) {
   const state = socket[kStateSymbol];
 
   state.handle.onmessage = onMessage;
-  // Todo: handle errors
+  state.handle.onerror = onError;
   state.handle.recvStart();
   state.receiving = true;
   state.bindState = BIND_STATE_BOUND;
@@ -154,18 +185,15 @@ function replaceHandle(self, newHandle) {
   const state = self[kStateSymbol];
   const oldHandle = state.handle;
 
-  // Set up the handle that we got from master.
+  // Set up the handle that we got from primary.
   newHandle.lookup = oldHandle.lookup;
   newHandle.bind = oldHandle.bind;
   newHandle.send = oldHandle.send;
   newHandle[owner_symbol] = self;
 
-  // Replace the existing handle by the handle we got from master.
+  // Replace the existing handle by the handle we got from primary.
   oldHandle.close();
   state.handle = newHandle;
-  // Check if the udp handle was connected and set the state accordingly
-  if (isConnected(self))
-    state.connectState = CONNECT_STATE_CONNECTED;
 }
 
 function bufferSize(self, size, buffer) {
@@ -180,10 +208,9 @@ function bufferSize(self, size, buffer) {
   return ret;
 }
 
-// Query master process to get the server handle and utilize it.
+// Query primary process to get the server handle and utilize it.
 function bindServerHandle(self, options, errCb) {
-  if (!cluster)
-    cluster = require('cluster');
+  const cluster = lazyLoadCluster();
 
   const state = self[kStateSymbol];
   cluster._getServer(self, options, (err, handle) => {
@@ -221,8 +248,8 @@ Socket.prototype.bind = function(port_, address_ /* , callback */) {
     }
 
     function onListening() {
-      removeListeners.call(this);
-      cb.call(this);
+      FunctionPrototypeCall(removeListeners, this);
+      FunctionPrototypeCall(cb, this);
     }
 
     this.on('error', removeListeners);
@@ -244,8 +271,7 @@ Socket.prototype.bind = function(port_, address_ /* , callback */) {
     const exclusive = !!port.exclusive;
     const state = this[kStateSymbol];
 
-    if (!cluster)
-      cluster = require('cluster');
+    const cluster = lazyLoadCluster();
 
     if (cluster.isWorker && !exclusive) {
       bindServerHandle(this, {
@@ -271,10 +297,6 @@ Socket.prototype.bind = function(port_, address_ /* , callback */) {
     if (err)
       throw errnoException(err, 'open');
 
-    // Check if the udp handle was connected and set the state accordingly
-    if (isConnected(this))
-      state.connectState = CONNECT_STATE_CONNECTED;
-
     startListening(this);
     return this;
   }
@@ -307,8 +329,7 @@ Socket.prototype.bind = function(port_, address_ /* , callback */) {
       return;
     }
 
-    if (!cluster)
-      cluster = require('cluster');
+    const cluster = lazyLoadCluster();
 
     let flags = 0;
     if (state.reuseAddr)
@@ -350,7 +371,7 @@ Socket.prototype.bind = function(port_, address_ /* , callback */) {
 };
 
 Socket.prototype.connect = function(port, address, callback) {
-  port = validatePort(port, 'Port', { allowZero: false });
+  port = validatePort(port, 'Port', false);
   if (typeof address === 'function') {
     callback = address;
     address = '';
@@ -370,11 +391,12 @@ Socket.prototype.connect = function(port, address, callback) {
     this.bind({ port: 0, exclusive: true }, null);
 
   if (state.bindState !== BIND_STATE_BOUND) {
-    enqueue(this, _connect.bind(this, port, address, callback));
+    enqueue(this, FunctionPrototypeBind(_connect, this,
+                                        port, address, callback));
     return;
   }
 
-  _connect.call(this, port, address, callback);
+  ReflectApply(_connect, this, [port, address, callback]);
 };
 
 
@@ -467,6 +489,13 @@ function sliceBuffer(buffer, offset, length) {
 
   offset = offset >>> 0;
   length = length >>> 0;
+  if (offset > buffer.byteLength) {
+    throw new ERR_BUFFER_OUT_OF_BOUNDS('offset');
+  }
+
+  if (offset + length > buffer.byteLength) {
+    throw new ERR_BUFFER_OUT_OF_BOUNDS('length');
+  }
 
   return Buffer.from(buffer.buffer, buffer.byteOffset + offset, length);
 }
@@ -496,23 +525,22 @@ function enqueue(self, toEnqueue) {
   // event handler that flushes the send queue after binding is done.
   if (state.queue === undefined) {
     state.queue = [];
-    self.once('error', onListenError);
+    self.once(EventEmitter.errorMonitor, onListenError);
     self.once('listening', onListenSuccess);
   }
-  state.queue.push(toEnqueue);
+  ArrayPrototypePush(state.queue, toEnqueue);
 }
 
 
 function onListenSuccess() {
-  this.removeListener('error', onListenError);
-  clearQueue.call(this);
+  this.removeListener(EventEmitter.errorMonitor, onListenError);
+  FunctionPrototypeCall(clearQueue, this);
 }
 
 
 function onListenError(err) {
   this.removeListener('listening', onListenSuccess);
   this[kStateSymbol].queue = undefined;
-  this.emit('error', new ERR_SOCKET_CANNOT_SEND());
 }
 
 
@@ -526,16 +554,6 @@ function clearQueue() {
     queueEntry();
 }
 
-function isConnected(self) {
-  try {
-    self.remoteAddress();
-    return true;
-  } catch {
-    return false;
-  }
-}
-
-
 // valid combinations
 // For connectionless sockets
 // send(buffer, offset, length, port, address, callback)
@@ -607,7 +625,7 @@ Socket.prototype.send = function(buffer,
   }
 
   if (!connected)
-    port = validatePort(port, 'Port', { allowZero: false });
+    port = validatePort(port, 'Port', false);
 
   // Normalize callback so it's either a function or undefined but not anything
   // else.
@@ -617,8 +635,8 @@ Socket.prototype.send = function(buffer,
   if (typeof address === 'function') {
     callback = address;
     address = undefined;
-  } else if (address && typeof address !== 'string') {
-    throw new ERR_INVALID_ARG_TYPE('address', ['string', 'falsy'], address);
+  } else if (address != null) {
+    validateString(address, 'address');
   }
 
   healthCheck(this);
@@ -627,12 +645,13 @@ Socket.prototype.send = function(buffer,
     this.bind({ port: 0, exclusive: true }, null);
 
   if (list.length === 0)
-    list.push(Buffer.alloc(0));
+    ArrayPrototypePush(list, Buffer.alloc(0));
 
   // If the socket hasn't been bound yet, push the outbound packet onto the
   // send queue and send after binding is complete.
   if (state.bindState !== BIND_STATE_BOUND) {
-    enqueue(this, this.send.bind(this, list, port, address, callback));
+    enqueue(this, FunctionPrototypeBind(this.send, this,
+                                        list, port, address, callback));
     return;
   }
 
@@ -714,7 +733,7 @@ Socket.prototype.close = function(callback) {
     this.on('close', callback);
 
   if (queue !== undefined) {
-    queue.push(this.close.bind(this));
+    ArrayPrototypePush(queue, FunctionPrototypeBind(this.close, this));
     return this;
   }
 
@@ -852,13 +871,8 @@ Socket.prototype.addSourceSpecificMembership = function(sourceAddress,
                                                         interfaceAddress) {
   healthCheck(this);
 
-  if (typeof sourceAddress !== 'string') {
-    throw new ERR_INVALID_ARG_TYPE('sourceAddress', 'string', sourceAddress);
-  }
-
-  if (typeof groupAddress !== 'string') {
-    throw new ERR_INVALID_ARG_TYPE('groupAddress', 'string', groupAddress);
-  }
+  validateString(sourceAddress, 'sourceAddress');
+  validateString(groupAddress, 'groupAddress');
 
   const err =
     this[kStateSymbol].handle.addSourceSpecificMembership(sourceAddress,
@@ -875,13 +889,8 @@ Socket.prototype.dropSourceSpecificMembership = function(sourceAddress,
                                                          interfaceAddress) {
   healthCheck(this);
 
-  if (typeof sourceAddress !== 'string') {
-    throw new ERR_INVALID_ARG_TYPE('sourceAddress', 'string', sourceAddress);
-  }
-
-  if (typeof groupAddress !== 'string') {
-    throw new ERR_INVALID_ARG_TYPE('groupAddress', 'string', groupAddress);
-  }
+  validateString(sourceAddress, 'sourceAddress');
+  validateString(groupAddress, 'groupAddress');
 
   const err =
     this[kStateSymbol].handle.dropSourceSpecificMembership(sourceAddress,
@@ -922,6 +931,12 @@ function onMessage(nread, handle, buf, rinfo) {
 }
 
 
+function onError(nread, handle, error) {
+  const self = handle[owner_symbol];
+  return self.emit('error', error);
+}
+
+
 Socket.prototype.ref = function() {
   const handle = this[kStateSymbol].handle;
 
@@ -961,9 +976,17 @@ Socket.prototype.getSendBufferSize = function() {
   return bufferSize(this, 0, SEND_BUFFER);
 };
 
+Socket.prototype.getSendQueueSize = function() {
+  return this[kStateSymbol].handle.getSendQueueSize();
+};
+
+Socket.prototype.getSendQueueCount = function() {
+  return this[kStateSymbol].handle.getSendQueueCount();
+};
 
 // Deprecated private APIs.
 ObjectDefineProperty(Socket.prototype, '_handle', {
+  __proto__: null,
   get: deprecate(function() {
     return this[kStateSymbol].handle;
   }, 'Socket.prototype._handle is deprecated', 'DEP0112'),
@@ -974,6 +997,7 @@ ObjectDefineProperty(Socket.prototype, '_handle', {
 
 
 ObjectDefineProperty(Socket.prototype, '_receiving', {
+  __proto__: null,
   get: deprecate(function() {
     return this[kStateSymbol].receiving;
   }, 'Socket.prototype._receiving is deprecated', 'DEP0112'),
@@ -984,6 +1008,7 @@ ObjectDefineProperty(Socket.prototype, '_receiving', {
 
 
 ObjectDefineProperty(Socket.prototype, '_bindState', {
+  __proto__: null,
   get: deprecate(function() {
     return this[kStateSymbol].bindState;
   }, 'Socket.prototype._bindState is deprecated', 'DEP0112'),
@@ -994,6 +1019,7 @@ ObjectDefineProperty(Socket.prototype, '_bindState', {
 
 
 ObjectDefineProperty(Socket.prototype, '_queue', {
+  __proto__: null,
   get: deprecate(function() {
     return this[kStateSymbol].queue;
   }, 'Socket.prototype._queue is deprecated', 'DEP0112'),
@@ -1004,6 +1030,7 @@ ObjectDefineProperty(Socket.prototype, '_queue', {
 
 
 ObjectDefineProperty(Socket.prototype, '_reuseAddr', {
+  __proto__: null,
   get: deprecate(function() {
     return this[kStateSymbol].reuseAddr;
   }, 'Socket.prototype._reuseAddr is deprecated', 'DEP0112'),
@@ -1026,6 +1053,7 @@ Socket.prototype._stopReceiving = deprecate(function() {
 // Legacy alias on the C++ wrapper object. This is not public API, so we may
 // want to runtime-deprecate it at some point. There's no hurry, though.
 ObjectDefineProperty(UDP.prototype, 'owner', {
+  __proto__: null,
   get() { return this[owner_symbol]; },
   set(v) { return this[owner_symbol] = v; }
 });
