@@ -770,7 +770,7 @@ const validateRmOptions = hideStackFrames((path, options, expectDir, cb) => {
   options = validateRmdirOptions(options, defaultRmOptions);
   validateBoolean(options.force, 'options.force');
 
-  lazyLoadFs().stat(path, (err, stats) => {
+  lazyLoadFs().lstat(path, (err, stats) => {
     if (err) {
       if (options.force && err.code === 'ENOENT') {
         return cb(null, options);
@@ -801,7 +801,7 @@ const validateRmOptionsSync = hideStackFrames((path, options, expectDir) => {
 
   if (!options.force || expectDir || !options.recursive) {
     const isDirectory = lazyLoadFs()
-      .statSync(path, { throwIfNoEntry: !options.force })?.isDirectory();
+      .lstatSync(path, { throwIfNoEntry: !options.force })?.isDirectory();
 
     if (expectDir && !isDirectory) {
       return false;
@@ -906,17 +906,15 @@ const validatePrimitiveStringAfterArrayBufferView = hideStackFrames((buffer, nam
 
 const validatePosition = hideStackFrames((position, name) => {
   if (typeof position === 'number') {
-    validateInteger(position, 'position');
+    validateInteger(position, name);
   } else if (typeof position === 'bigint') {
     if (!(position >= -(2n ** 63n) && position <= 2n ** 63n - 1n)) {
-      throw new ERR_OUT_OF_RANGE('position',
+      throw new ERR_OUT_OF_RANGE(name,
                                  `>= ${-(2n ** 63n)} && <= ${2n ** 63n - 1n}`,
                                  position);
     }
   } else {
-    throw new ERR_INVALID_ARG_TYPE('position',
-                                   ['integer', 'bigint'],
-                                   position);
+    throw new ERR_INVALID_ARG_TYPE(name, ['integer', 'bigint'], position);
   }
 });
 
