@@ -7,6 +7,8 @@ const { toNamespacedPath } = require('path');
 
 const cache = new SafeMap();
 
+let manifest;
+
 /**
  *
  * @param {string} jsonPath
@@ -16,16 +18,18 @@ function read(jsonPath) {
     return cache.get(jsonPath);
   }
 
-  const [string, containsKeys] = internalModuleReadJSON(
+  const { 0: string, 1: containsKeys } = internalModuleReadJSON(
     toNamespacedPath(jsonPath)
   );
   const result = { string, containsKeys };
   const { getOptionValue } = require('internal/options');
   if (string !== undefined) {
-    const manifest = getOptionValue('--experimental-policy') ?
-      require('internal/process/policy').manifest :
-      null;
-    if (manifest) {
+    if (manifest === undefined) {
+      manifest = getOptionValue('--experimental-policy') ?
+        require('internal/process/policy').manifest :
+        null;
+    }
+    if (manifest !== null) {
       const jsonURL = pathToFileURL(jsonPath);
       manifest.assertIntegrity(jsonURL, string);
     }
