@@ -21,16 +21,24 @@
 
 'use strict';
 
-const { toASCII } = process.binding('config').hasIntl ?
-  process.binding('icu') : require('punycode');
+const {
+  Int8Array,
+  ObjectCreate,
+  ObjectKeys,
+  SafeSet,
+  StringPrototypeCharCodeAt,
+  decodeURIComponent,
+} = primordials;
 
-const { hexTable } = require('internal/querystring');
-
-const { SafeSet } = require('internal/safe_globals');
+const { toASCII } = require('internal/idna');
+const { encodeStr, hexTable } = require('internal/querystring');
+const querystring = require('querystring');
 
 const {
-  ERR_INVALID_ARG_TYPE
+  ERR_INVALID_ARG_TYPE,
+  ERR_INVALID_URL,
 } = require('internal/errors').codes;
+const { validateString } = require('internal/validators');
 
 // This ensures setURLConstructor() is called before the native
 // URL::ToObject() method is used.
@@ -42,10 +50,10 @@ const {
   URLSearchParams,
   domainToASCII,
   domainToUnicode,
+  fileURLToPath,
   formatSymbol,
-  encodeStr,
   pathToFileURL,
-  fileURLToPath
+  urlToHttpOptions,
 } = require('internal/url');
 
 // Original url.parse() API
@@ -77,17 +85,17 @@ const hostPattern = /^\/\/[^@/]+@[^@/]+/;
 const simplePathPattern = /^(\/\/?(?!\/)[^?\s]*)(\?[^\s]*)?$/;
 
 const hostnameMaxLen = 255;
-// protocols that can allow "unsafe" and "unwise" chars.
+// Protocols that can allow "unsafe" and "unwise" chars.
 const unsafeProtocol = new SafeSet([
   'javascript',
-  'javascript:'
+  'javascript:',
 ]);
-// protocols that never have a hostname.
+// Protocols that never have a hostname.
 const hostlessProtocol = new SafeSet([
   'javascript',
-  'javascript:'
+  'javascript:',
 ]);
-// protocols that always contain a // bit.
+// Protocols that always contain a // bit.
 const slashedProtocol = new SafeSet([
   'http',
   'http:',
@@ -98,14 +106,17 @@ const slashedProtocol = new SafeSet([
   'gopher',
   'gopher:',
   'file',
-  'file:'
+  'file:',
+  'ws',
+  'ws:',
+  'wss',
+  'wss:',
 ]);
 const {
   CHAR_SPACE,
   CHAR_TAB,
   CHAR_CARRIAGE_RETURN,
   CHAR_LINE_FEED,
-  CHAR_FORM_FEED,
   CHAR_NO_BREAK_SPACE,
   CHAR_ZERO_WIDTH_NOBREAK_SPACE,
   CHAR_HASH,
@@ -117,16 +128,6 @@ const {
   CHAR_LEFT_CURLY_BRACKET,
   CHAR_RIGHT_CURLY_BRACKET,
   CHAR_QUESTION_MARK,
-  CHAR_LOWERCASE_A,
-  CHAR_LOWERCASE_Z,
-  CHAR_UPPERCASE_A,
-  CHAR_UPPERCASE_Z,
-  CHAR_DOT,
-  CHAR_0,
-  CHAR_9,
-  CHAR_HYPHEN_MINUS,
-  CHAR_PLUS,
-  CHAR_UNDERSCORE,
   CHAR_DOUBLE_QUOTE,
   CHAR_SINGLE_QUOTE,
   CHAR_PERCENT,
@@ -136,42 +137,58 @@ const {
   CHAR_GRAVE_ACCENT,
   CHAR_VERTICAL_LINE,
   CHAR_AT,
+  CHAR_COLON,
 } = require('internal/constants');
 
-// Lazy loaded for startup performance.
-let querystring;
-
 function urlParse(url, parseQueryString, slashesDenoteHost) {
   if (url instanceof Url) return url;
 
-  var urlObject = new Url();
+  const urlObject = new Url();
   urlObject.parse(url, parseQueryString, slashesDenoteHost);
   return urlObject;
 }
 
+function isIpv6Hostname(hostname) {
+  return (
+    StringPrototypeCharCodeAt(hostname, 0) === CHAR_LEFT_SQUARE_BRACKET &&
+    StringPrototypeCharCodeAt(hostname, hostname.length - 1) ===
+    CHAR_RIGHT_SQUARE_BRACKET
+  );
+}
+
+// This prevents some common spoofing bugs due to our use of IDNA toASCII. For
+// compatibility, the set of characters we use here is the *intersection* of
+// "forbidden host code point" in the WHATWG URL Standard [1] and the
+// characters in the host parsing loop in Url.prototype.parse, with the
+// following additions:
+//
+// - ':' since this could cause a "protocol spoofing" bug
+// - '@' since this could cause parts of the hostname to be confused with auth
+// - '[' and ']' since this could cause a non-IPv6 hostname to be interpreted
+//   as IPv6 by isIpv6Hostname above
+//
+// [1]: https://url.spec.whatwg.org/#forbidden-host-code-point
+const forbiddenHostChars = /[\0\t\n\r #%/:<>?@[\\\]^|]/;
+// For IPv6, permit '[', ']', and ':'.
+const forbiddenHostCharsIpv6 = /[\0\t\n\r #%/<>?@\\^|]/;
+
 Url.prototype.parse = function parse(url, parseQueryString, slashesDenoteHost) {
-  if (typeof url !== 'string') {
-    throw new ERR_INVALID_ARG_TYPE('url', 'string', url);
-  }
+  validateString(url, 'url');
 
   // Copy chrome, IE, opera backslash-handling behavior.
   // Back slashes before the query string get converted to forward slashes
   // See: https://code.google.com/p/chromium/issues/detail?id=25916
-  var hasHash = false;
-  var start = -1;
-  var end = -1;
-  var rest = '';
-  var lastPos = 0;
-  var i = 0;
-  for (var inWs = false, split = false; i < url.length; ++i) {
+  let hasHash = false;
+  let hasAt = false;
+  let start = -1;
+  let end = -1;
+  let rest = '';
+  let lastPos = 0;
+  for (let i = 0, inWs = false, split = false; i < url.length; ++i) {
     const code = url.charCodeAt(i);
 
     // Find first and last non-whitespace characters for trimming
-    const isWs = code === CHAR_SPACE ||
-                 code === CHAR_TAB ||
-                 code === CHAR_CARRIAGE_RETURN ||
-                 code === CHAR_LINE_FEED ||
-                 code === CHAR_FORM_FEED ||
+    const isWs = code < 33 ||
                  code === CHAR_NO_BREAK_SPACE ||
                  code === CHAR_ZERO_WIDTH_NOBREAK_SPACE;
     if (start === -1) {
@@ -191,6 +208,9 @@ Url.prototype.parse = function parse(url, parseQueryString, slashesDenoteHost) {
     // Only convert backslashes while we haven't seen a split character
     if (!split) {
       switch (code) {
+        case CHAR_AT:
+          hasAt = true;
+          break;
         case CHAR_HASH:
           hasHash = true;
         // Fall through
@@ -231,7 +251,7 @@ Url.prototype.parse = function parse(url, parseQueryString, slashesDenoteHost) {
     }
   }
 
-  if (!slashesDenoteHost && !hasHash) {
+  if (!slashesDenoteHost && !hasHash && !hasAt) {
     // Try fast path regexp
     const simplePath = simplePathPattern.exec(rest);
     if (simplePath) {
@@ -241,34 +261,35 @@ Url.prototype.parse = function parse(url, parseQueryString, slashesDenoteHost) {
       if (simplePath[2]) {
         this.search = simplePath[2];
         if (parseQueryString) {
-          if (querystring === undefined) querystring = require('querystring');
           this.query = querystring.parse(this.search.slice(1));
         } else {
           this.query = this.search.slice(1);
         }
       } else if (parseQueryString) {
         this.search = null;
-        this.query = Object.create(null);
+        this.query = ObjectCreate(null);
       }
       return this;
     }
   }
 
-  var proto = protocolPattern.exec(rest);
+  let proto = protocolPattern.exec(rest);
+  let lowerProto;
   if (proto) {
     proto = proto[0];
-    var lowerProto = proto.toLowerCase();
+    lowerProto = proto.toLowerCase();
     this.protocol = lowerProto;
     rest = rest.slice(proto.length);
   }
 
-  // figure out if it's got a host
+  // Figure out if it's got a host
   // user@server is *always* interpreted as a hostname, and url
   // resolution will treat //foo/bar as host=foo,path=bar because that's
   // how the browser resolves relative URLs.
+  let slashes;
   if (slashesDenoteHost || proto || hostPattern.test(rest)) {
-    var slashes = rest.charCodeAt(0) === CHAR_FORWARD_SLASH &&
-                  rest.charCodeAt(1) === CHAR_FORWARD_SLASH;
+    slashes = rest.charCodeAt(0) === CHAR_FORWARD_SLASH &&
+              rest.charCodeAt(1) === CHAR_FORWARD_SLASH;
     if (slashes && !(proto && hostlessProtocol.has(lowerProto))) {
       rest = rest.slice(2);
       this.slashes = true;
@@ -290,14 +311,18 @@ Url.prototype.parse = function parse(url, parseQueryString, slashesDenoteHost) {
     // http://a@b@c/ => user:a@b host:c
     // http://a@b?@c => user:a host:b path:/?@c
 
-    var hostEnd = -1;
-    var atSign = -1;
-    var nonHost = -1;
-    for (i = 0; i < rest.length; ++i) {
+    let hostEnd = -1;
+    let atSign = -1;
+    let nonHost = -1;
+    for (let i = 0; i < rest.length; ++i) {
       switch (rest.charCodeAt(i)) {
         case CHAR_TAB:
         case CHAR_LINE_FEED:
         case CHAR_CARRIAGE_RETURN:
+          // WHATWG URL removes tabs, newlines, and carriage returns. Let's do that too.
+          rest = rest.slice(0, i) + rest.slice(i + 1);
+          i -= 1;
+          break;
         case CHAR_SPACE:
         case CHAR_DOUBLE_QUOTE:
         case CHAR_PERCENT:
@@ -349,45 +374,61 @@ Url.prototype.parse = function parse(url, parseQueryString, slashesDenoteHost) {
     // pull out port.
     this.parseHost();
 
-    // we've indicated that there is a hostname,
+    // We've indicated that there is a hostname,
     // so even if it's empty, it has to be present.
     if (typeof this.hostname !== 'string')
       this.hostname = '';
 
-    var hostname = this.hostname;
+    const hostname = this.hostname;
 
-    // if hostname begins with [ and ends with ]
+    // If hostname begins with [ and ends with ]
     // assume that it's an IPv6 address.
-    var ipv6Hostname = hostname.charCodeAt(0) === CHAR_LEFT_SQUARE_BRACKET &&
-      hostname.charCodeAt(hostname.length - 1) === CHAR_RIGHT_SQUARE_BRACKET;
+    const ipv6Hostname = isIpv6Hostname(hostname);
 
     // validate a little.
     if (!ipv6Hostname) {
-      const result = validateHostname(this, rest, hostname);
-      if (result !== undefined)
-        rest = result;
+      rest = getHostname(this, rest, hostname);
     }
 
     if (this.hostname.length > hostnameMaxLen) {
       this.hostname = '';
     } else {
-      // hostnames are always lower case.
+      // Hostnames are always lower case.
       this.hostname = this.hostname.toLowerCase();
     }
 
-    if (!ipv6Hostname) {
-      // IDNA Support: Returns a punycoded representation of "domain".
-      // It only converts parts of the domain name that
-      // have non-ASCII characters, i.e. it doesn't matter if
-      // you call it with a domain that already is ASCII-only.
+    if (this.hostname !== '') {
+      if (ipv6Hostname) {
+        if (forbiddenHostCharsIpv6.test(this.hostname)) {
+          throw new ERR_INVALID_URL(url);
+        }
+      } else {
+        // IDNA Support: Returns a punycoded representation of "domain".
+        // It only converts parts of the domain name that
+        // have non-ASCII characters, i.e. it doesn't matter if
+        // you call it with a domain that already is ASCII-only.
 
-      // Use lenient mode (`true`) to try to support even non-compliant
-      // URLs.
-      this.hostname = toASCII(this.hostname, true);
+        // Use lenient mode (`true`) to try to support even non-compliant
+        // URLs.
+        this.hostname = toASCII(this.hostname, true);
+
+        // Prevent two potential routes of hostname spoofing.
+        // 1. If this.hostname is empty, it must have become empty due to toASCII
+        //    since we checked this.hostname above.
+        // 2. If any of forbiddenHostChars appears in this.hostname, it must have
+        //    also gotten in due to toASCII. This is since getHostname would have
+        //    filtered them out otherwise.
+        // Rather than trying to correct this by moving the non-host part into
+        // the pathname as we've done in getHostname, throw an exception to
+        // convey the severity of this issue.
+        if (this.hostname === '' || forbiddenHostChars.test(this.hostname)) {
+          throw new ERR_INVALID_URL(url);
+        }
+      }
     }
 
-    var p = this.port ? ':' + this.port : '';
-    var h = this.hostname || '';
+    const p = this.port ? ':' + this.port : '';
+    const h = this.hostname || '';
     this.host = h + p;
 
     // strip [ and ] from the hostname
@@ -400,8 +441,8 @@ Url.prototype.parse = function parse(url, parseQueryString, slashesDenoteHost) {
     }
   }
 
-  // now rest is set to the post-host stuff.
-  // chop off any delim chars.
+  // Now rest is set to the post-host stuff.
+  // Chop off any delim chars.
   if (!unsafeProtocol.has(lowerProto)) {
     // First, make 100% sure that any "autoEscape" chars get
     // escaped, even if encodeURIComponent doesn't think they
@@ -409,9 +450,9 @@ Url.prototype.parse = function parse(url, parseQueryString, slashesDenoteHost) {
     rest = autoEscapeStr(rest);
   }
 
-  var questionIdx = -1;
-  var hashIdx = -1;
-  for (i = 0; i < rest.length; ++i) {
+  let questionIdx = -1;
+  let hashIdx = -1;
+  for (let i = 0; i < rest.length; ++i) {
     const code = rest.charCodeAt(i);
     if (code === CHAR_HASH) {
       this.hash = rest.slice(i);
@@ -431,13 +472,12 @@ Url.prototype.parse = function parse(url, parseQueryString, slashesDenoteHost) {
       this.query = rest.slice(questionIdx + 1, hashIdx);
     }
     if (parseQueryString) {
-      if (querystring === undefined) querystring = require('querystring');
       this.query = querystring.parse(this.query);
     }
   } else if (parseQueryString) {
-    // no query string, but parseQueryString still requested
+    // No query string, but parseQueryString still requested
     this.search = null;
-    this.query = Object.create(null);
+    this.query = ObjectCreate(null);
   }
 
   const useQuestionIdx =
@@ -454,36 +494,33 @@ Url.prototype.parse = function parse(url, parseQueryString, slashesDenoteHost) {
     this.pathname = '/';
   }
 
-  // to support http.request
+  // To support http.request
   if (this.pathname || this.search) {
     const p = this.pathname || '';
     const s = this.search || '';
     this.path = p + s;
   }
 
-  // finally, reconstruct the href based on what has been validated.
+  // Finally, reconstruct the href based on what has been validated.
   this.href = this.format();
   return this;
 };
 
-function validateHostname(self, rest, hostname) {
-  for (var i = 0; i < hostname.length; ++i) {
+function getHostname(self, rest, hostname) {
+  for (let i = 0; i < hostname.length; ++i) {
     const code = hostname.charCodeAt(i);
-    const isValid = (code >= CHAR_LOWERCASE_A && code <= CHAR_LOWERCASE_Z) ||
-                    code === CHAR_DOT ||
-                    (code >= CHAR_UPPERCASE_A && code <= CHAR_UPPERCASE_Z) ||
-                    (code >= CHAR_0 && code <= CHAR_9) ||
-                    code === CHAR_HYPHEN_MINUS ||
-                    code === CHAR_PLUS ||
-                    code === CHAR_UNDERSCORE ||
-                    code > 127;
+    const isValid = (code !== CHAR_FORWARD_SLASH &&
+                     code !== CHAR_BACKWARD_SLASH &&
+                     code !== CHAR_HASH &&
+                     code !== CHAR_QUESTION_MARK &&
+                     code !== CHAR_COLON);
 
-    // Invalid host character
     if (!isValid) {
       self.hostname = hostname.slice(0, i);
-      return '/' + hostname.slice(i) + rest;
+      return `/${hostname.slice(i)}${rest}`;
     }
   }
+  return rest;
 }
 
 // Escaped characters. Use empty strings to fill up unused entries.
@@ -501,18 +538,18 @@ const escapedCodes = [
   /* 90 - 99 */ '', '', '%5C', '', '%5E', '', '%60', '', '', '',
   /* 100 - 109 */ '', '', '', '', '', '', '', '', '', '',
   /* 110 - 119 */ '', '', '', '', '', '', '', '', '', '',
-  /* 120 - 125 */ '', '', '', '%7B', '%7C', '%7D'
+  /* 120 - 125 */ '', '', '', '%7B', '%7C', '%7D',
 ];
 
 // Automatically escape all delimiters and unwise characters from RFC 2396.
 // Also escape single quotes in case of an XSS attack.
 // Return the escaped string.
 function autoEscapeStr(rest) {
-  var escaped = '';
-  var lastEscapedPos = 0;
-  for (var i = 0; i < rest.length; ++i) {
+  let escaped = '';
+  let lastEscapedPos = 0;
+  for (let i = 0; i < rest.length; ++i) {
     // `escaped` contains substring up to the last escaped character.
-    var escapedChar = escapedCodes[rest.charCodeAt(i)];
+    const escapedChar = escapedCodes[rest.charCodeAt(i)];
     if (escapedChar) {
       // Concat if there are ordinary characters in the middle.
       if (i > lastEscapedPos)
@@ -531,9 +568,9 @@ function autoEscapeStr(rest) {
   return escaped;
 }
 
-// format a parsed object into a url string
+// Format a parsed object into a url string
 function urlFormat(urlObject, options) {
-  // ensure it's an object, and not a string url.
+  // Ensure it's an object, and not a string url.
   // If it's an object, this is a no-op.
   // this way, you can call urlParse() on strings
   // to clean up potentially wonky urls.
@@ -543,7 +580,7 @@ function urlFormat(urlObject, options) {
     throw new ERR_INVALID_ARG_TYPE('urlObject',
                                    ['Object', 'string'], urlObject);
   } else if (!(urlObject instanceof Url)) {
-    var format = urlObject[formatSymbol];
+    const format = urlObject[formatSymbol];
     return format ?
       format.call(urlObject, options) :
       Url.prototype.format.call(urlObject);
@@ -557,7 +594,7 @@ function urlFormat(urlObject, options) {
 // digits
 // alpha (uppercase)
 // alpha (lowercase)
-const noEscapeAuth = [
+const noEscapeAuth = new Int8Array([
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, // 0x00 - 0x0F
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, // 0x10 - 0x1F
   0, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 1, 1, 0, // 0x20 - 0x2F
@@ -565,29 +602,29 @@ const noEscapeAuth = [
   0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, // 0x40 - 0x4F
   1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1, // 0x50 - 0x5F
   0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, // 0x60 - 0x6F
-  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 1, 0  // 0x70 - 0x7F
-];
+  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 1, 0,  // 0x70 - 0x7F
+]);
 
 Url.prototype.format = function format() {
-  var auth = this.auth || '';
+  let auth = this.auth || '';
   if (auth) {
     auth = encodeStr(auth, noEscapeAuth, hexTable);
     auth += '@';
   }
 
-  var protocol = this.protocol || '';
-  var pathname = this.pathname || '';
-  var hash = this.hash || '';
-  var host = '';
-  var query = '';
+  let protocol = this.protocol || '';
+  let pathname = this.pathname || '';
+  let hash = this.hash || '';
+  let host = '';
+  let query = '';
 
   if (this.host) {
     host = auth + this.host;
   } else if (this.hostname) {
     host = auth + (
-      this.hostname.indexOf(':') === -1 ?
-        this.hostname :
-        '[' + this.hostname + ']'
+      this.hostname.includes(':') && !isIpv6Hostname(this.hostname) ?
+        '[' + this.hostname + ']' :
+        this.hostname
     );
     if (this.port) {
       host += ':' + this.port;
@@ -595,18 +632,17 @@ Url.prototype.format = function format() {
   }
 
   if (this.query !== null && typeof this.query === 'object') {
-    if (querystring === undefined) querystring = require('querystring');
     query = querystring.stringify(this.query);
   }
 
-  var search = this.search || (query && ('?' + query)) || '';
+  let search = this.search || (query && ('?' + query)) || '';
 
   if (protocol && protocol.charCodeAt(protocol.length - 1) !== 58/* : */)
     protocol += ':';
 
-  var newPathname = '';
-  var lastPos = 0;
-  for (var i = 0; i < pathname.length; ++i) {
+  let newPathname = '';
+  let lastPos = 0;
+  for (let i = 0; i < pathname.length; ++i) {
     switch (pathname.charCodeAt(i)) {
       case CHAR_HASH:
         if (i - lastPos > 0)
@@ -629,7 +665,7 @@ Url.prototype.format = function format() {
       pathname = newPathname;
   }
 
-  // only the slashedProtocols get the //.  Not mailto:, xmpp:, etc.
+  // Only the slashedProtocols get the //.  Not mailto:, xmpp:, etc.
   // unless they had them to begin with.
   if (this.slashes || slashedProtocol.has(protocol)) {
     if (this.slashes || host) {
@@ -670,34 +706,34 @@ function urlResolveObject(source, relative) {
 
 Url.prototype.resolveObject = function resolveObject(relative) {
   if (typeof relative === 'string') {
-    var rel = new Url();
+    const rel = new Url();
     rel.parse(relative, false, true);
     relative = rel;
   }
 
-  var result = new Url();
-  var tkeys = Object.keys(this);
-  for (var tk = 0; tk < tkeys.length; tk++) {
-    var tkey = tkeys[tk];
+  const result = new Url();
+  const tkeys = ObjectKeys(this);
+  for (let tk = 0; tk < tkeys.length; tk++) {
+    const tkey = tkeys[tk];
     result[tkey] = this[tkey];
   }
 
-  // hash is always overridden, no matter what.
+  // Hash is always overridden, no matter what.
   // even href="" will remove it.
   result.hash = relative.hash;
 
-  // if the relative url is empty, then there's nothing left to do here.
+  // If the relative url is empty, then there's nothing left to do here.
   if (relative.href === '') {
     result.href = result.format();
     return result;
   }
 
-  // hrefs like //foo/bar always cut to the protocol.
+  // Hrefs like //foo/bar always cut to the protocol.
   if (relative.slashes && !relative.protocol) {
-    // take everything except the protocol from relative
-    var rkeys = Object.keys(relative);
-    for (var rk = 0; rk < rkeys.length; rk++) {
-      var rkey = rkeys[rk];
+    // Take everything except the protocol from relative
+    const rkeys = ObjectKeys(relative);
+    for (let rk = 0; rk < rkeys.length; rk++) {
+      const rkey = rkeys[rk];
       if (rkey !== 'protocol')
         result[rkey] = relative[rkey];
     }
@@ -713,7 +749,7 @@ Url.prototype.resolveObject = function resolveObject(relative) {
   }
 
   if (relative.protocol && relative.protocol !== result.protocol) {
-    // if it's a known url protocol, then changing
+    // If it's a known url protocol, then changing
     // the protocol does weird things
     // first, if it's not file:, then we MUST have a host,
     // and if there was a path
@@ -722,9 +758,9 @@ Url.prototype.resolveObject = function resolveObject(relative) {
     // because that's known to be hostless.
     // anything else is assumed to be absolute.
     if (!slashedProtocol.has(relative.protocol)) {
-      var keys = Object.keys(relative);
-      for (var v = 0; v < keys.length; v++) {
-        var k = keys[v];
+      const keys = ObjectKeys(relative);
+      for (let v = 0; v < keys.length; v++) {
+        const k = keys[v];
         result[k] = relative[k];
       }
       result.href = result.format();
@@ -751,10 +787,10 @@ Url.prototype.resolveObject = function resolveObject(relative) {
     result.auth = relative.auth;
     result.hostname = relative.hostname || relative.host;
     result.port = relative.port;
-    // to support http.request
+    // To support http.request
     if (result.pathname || result.search) {
-      var p = result.pathname || '';
-      var s = result.search || '';
+      const p = result.pathname || '';
+      const s = result.search || '';
       result.path = p + s;
     }
     result.slashes = result.slashes || relative.slashes;
@@ -762,19 +798,19 @@ Url.prototype.resolveObject = function resolveObject(relative) {
     return result;
   }
 
-  var isSourceAbs = (result.pathname && result.pathname.charAt(0) === '/');
-  var isRelAbs = (
-    relative.host || relative.pathname && relative.pathname.charAt(0) === '/'
+  const isSourceAbs = (result.pathname && result.pathname.charAt(0) === '/');
+  const isRelAbs = (
+    relative.host || (relative.pathname && relative.pathname.charAt(0) === '/')
   );
-  var mustEndAbs = (isRelAbs || isSourceAbs ||
+  let mustEndAbs = (isRelAbs || isSourceAbs ||
                     (result.host && relative.pathname));
-  var removeAllDots = mustEndAbs;
-  var srcPath = result.pathname && result.pathname.split('/') || [];
-  var relPath = relative.pathname && relative.pathname.split('/') || [];
-  var noLeadingSlashes = result.protocol &&
+  const removeAllDots = mustEndAbs;
+  let srcPath = (result.pathname && result.pathname.split('/')) || [];
+  const relPath = (relative.pathname && relative.pathname.split('/')) || [];
+  const noLeadingSlashes = result.protocol &&
       !slashedProtocol.has(result.protocol);
 
-  // if the url is a non-slashed url, then relative
+  // If the url is a non-slashed url, then relative
   // links like ../.. should be able
   // to crawl up to the hostname, as well.  This is strange.
   // result.protocol has already been set by now.
@@ -814,7 +850,7 @@ Url.prototype.resolveObject = function resolveObject(relative) {
     result.search = relative.search;
     result.query = relative.query;
     srcPath = relPath;
-    // fall through to the dot-handling below.
+    // Fall through to the dot-handling below.
   } else if (relPath.length) {
     // it's relative
     // throw away the existing file, and take the new path instead.
@@ -824,7 +860,7 @@ Url.prototype.resolveObject = function resolveObject(relative) {
     result.search = relative.search;
     result.query = relative.query;
   } else if (relative.search !== null && relative.search !== undefined) {
-    // just pull out the search.
+    // Just pull out the search.
     // like href='?foo'.
     // Put this after the other two cases because it simplifies the booleans
     if (noLeadingSlashes) {
@@ -851,8 +887,7 @@ Url.prototype.resolveObject = function resolveObject(relative) {
   }
 
   if (!srcPath.length) {
-    // no path at all.  easy.
-    // we've already handled the other stuff above.
+    // No path at all. All other things were already handled above.
     result.pathname = null;
     // To support http.request
     if (result.search) {
@@ -864,18 +899,18 @@ Url.prototype.resolveObject = function resolveObject(relative) {
     return result;
   }
 
-  // if a url ENDs in . or .., then it must get a trailing slash.
+  // If a url ENDs in . or .., then it must get a trailing slash.
   // however, if it ends in anything else non-slashy,
   // then it must NOT get a trailing slash.
-  var last = srcPath.slice(-1)[0];
-  var hasTrailingSlash = (
-    (result.host || relative.host || srcPath.length > 1) &&
-    (last === '.' || last === '..') || last === '');
+  let last = srcPath.slice(-1)[0];
+  const hasTrailingSlash = (
+    ((result.host || relative.host || srcPath.length > 1) &&
+    (last === '.' || last === '..')) || last === '');
 
-  // strip single dots, resolve double dots to parent dir
+  // Strip single dots, resolve double dots to parent dir
   // if the path tries to go above the root, `up` ends up > 0
-  var up = 0;
-  for (var i = srcPath.length - 1; i >= 0; i--) {
+  let up = 0;
+  for (let i = srcPath.length - 1; i >= 0; i--) {
     last = srcPath[i];
     if (last === '.') {
       spliceOne(srcPath, i);
@@ -888,7 +923,7 @@ Url.prototype.resolveObject = function resolveObject(relative) {
     }
   }
 
-  // if the path is allowed to go above the root, restore leading ..s
+  // If the path is allowed to go above the root, restore leading ..s
   if (!mustEndAbs && !removeAllDots) {
     while (up--) {
       srcPath.unshift('..');
@@ -904,7 +939,7 @@ Url.prototype.resolveObject = function resolveObject(relative) {
     srcPath.push('');
   }
 
-  var isAbsolute = srcPath[0] === '' ||
+  const isAbsolute = srcPath[0] === '' ||
       (srcPath[0] && srcPath[0].charAt(0) === '/');
 
   // put the host back
@@ -947,8 +982,8 @@ Url.prototype.resolveObject = function resolveObject(relative) {
 };
 
 Url.prototype.parseHost = function parseHost() {
-  var host = this.host;
-  var port = portPattern.exec(host);
+  let host = this.host;
+  let port = portPattern.exec(host);
   if (port) {
     port = port[0];
     if (port !== ':') {
@@ -975,5 +1010,6 @@ module.exports = {
 
   // Utilities
   pathToFileURL,
-  fileURLToPath
+  fileURLToPath,
+  urlToHttpOptions,
 };
